//+------------------------------------------------------------------+
//|                                                         ct_1.mq5 |
//|                                  Copyright 2022, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property                           copyright "Copyright 2022, MetaQuotes Ltd."
#property                           link      "https://www.mql5.com"
#property                           version   "1.00"
#property                           script_show_inputs
//+------------------------------------------------------------------+
//| ELEMENT CLASS                                                    |
//+------------------------------------------------------------------+
class CElement
   {
      protected:
      
      int                           cardinal;
      vector                        element;
      
      public:
      
      bool                          Cardinality(int Value) { if(Value>=0 && Value<INT_MAX) { cardinal=Value; element.Init(cardinal); return(true); } return(false); }
      int                           Cardinality() { return(cardinal); }
      
      double                        Get(int Index) { if(Index>=0 && Index<Cardinality()) { return(element[Index]); } return(EMPTY_VALUE); }
      bool                          Set(int Index,double Value) { if(Index>=0 && Index<Cardinality()) { element[Index]=Value; return(true); } return(false); }
      
                                    CElement(void)
                                    {
                                       Cardinality(0);
                                    };
                                    ~CElement(void) {};
   };
//+------------------------------------------------------------------+
//| DOMAIN CLASS                                                     |
//+------------------------------------------------------------------+
class CDomain
   {
      protected:
      
      int                           cardinal;
      CElement                      elements[];
      
      public:
      
      bool                          Cardinality(int Value) { if(Value>=0 && Value<INT_MAX) { cardinal=Value; ArrayResize(elements,cardinal); return(true); } return(false); }
      int                           Cardinality() { return(cardinal); }
      
      bool                          Get(int Index,CElement &Element) { if(Index>=0 && Index<Cardinality()) { Element=elements[Index]; return(true); } return(false); }
      bool                          Set(int Index,CElement &Value,bool IsNew=false) { if(Index>=0 && Index<Cardinality()) { if(!IsNew||New(Value)<0) { elements[Index]=Value; return(true); }} return(false); }
      
      //only unique elements allowed
      int                           New(CElement &Value)
                                    {
                                       bool _new=-1;
                                       //
                                       for(int o=0; o<cardinal; o++)
                                       {
                                          if(ElementMatch(Value,elements[o]))
                                          {
                                             _new=o;
                                             break;
                                          }
                                       }
                                       
                                       return(_new);
                                    }
      
                                    CDomain(void)
                                    {
                                       Cardinality(0);
                                    };
                                    ~CDomain(void) {};
   };
//+------------------------------------------------------------------+
//| MORPHISM CLASS                                                   |
//+------------------------------------------------------------------+
class CMorphism
   {
      public:
      
      int                           domain;
      int                           codomain;
      
      CElement                      morphism;
      
      bool                          Morph(CDomain &D,CDomain &C,CElement &DE,CElement &CE,bool Add=true)
                                    {
                                       int _d=D.New(DE),_c=C.New(CE);
                                       //
                                       if(_d>=0 && _c>=0)
                                       {
                                          if(DE.Cardinality()==CE.Cardinality())
                                          {
                                             domain=_d;
                                             codomain=_c;
                                             
                                             morphism.Cardinality(DE.Cardinality());
                                             
                                             if(Add)
                                             {
                                                for(int c=0;c<morphism.Cardinality();c++)
                                                {
                                                   morphism.Set(c,CE.Get(c)-DE.Get(c));
                                                }
                                             }
                                             else
                                             {
                                                for(int c=0;c<morphism.Cardinality();c++)
                                                {
                                                   if(DE.Get(c)!=0.0){ morphism.Set(c,CE.Get(c)/DE.Get(c)); }
                                                }
                                             }
                                          }
                                       }
                                       
                                       return(false);
                                    }
      
      
                                    CMorphism(void){ domain=-1; codomain=-1; };
                                    ~CMorphism(void){};
   };
//+------------------------------------------------------------------+
//| HOMO-MORPHISM CLASS                                              |
//+------------------------------------------------------------------+
class CHomomorphism
   {
      protected:
      
      int                           cardinal;
      
      CMorphism                     morphism[];
      
      public:
      
      bool                          init;
      
      CDomain                       domain;
      CDomain                       codomain;
      
      bool                          Cardinality(int DomainIndex,int CodomainIndex,bool Add=true)
                                    { 
                                       bool _morphed=true; 
                                       
                                       if
                                       (
                                       !init
                                       )
                                       {
                                          _morphed=false; return(_morphed); 
                                       }
                                       
                                       if
                                       (
                                       DomainIndex<0 || DomainIndex>=domain.Cardinality() ||
                                       CodomainIndex<0 || CodomainIndex>=codomain.Cardinality()
                                       )
                                       {
                                          _morphed=false; return(_morphed); 
                                       }
                                       
                                       for(int m=0;m<cardinal;m++)
                                       {
                                          if(DomainIndex==morphism[m].domain)
                                          {
                                             _morphed=false; break;
                                          }
                                       } 
                                       
                                       if(_morphed)
                                       {
                                          cardinal++;
                                          ArrayResize(morphism,cardinal);
                                          CElement _de,_ce;
                                          if(domain.Get(DomainIndex,_de) && codomain.Get(CodomainIndex,_ce))
                                          {
                                             morphism[cardinal-1].Morph(domain,codomain,_de,_ce,Add);
                                          }
                                       }
                                       
                                       return(_morphed); 
                                    };
      
      int                           Cardinality()
                                    {
                                       return(cardinal);
                                    };
                                    
      bool                          Get(int Index,CMorphism &Morphism)
                                    {
                                       if(Index>=0 && Index<Cardinality())
                                       {
                                          Morphism=morphism[Index];
                                          
                                          return(true);
                                       }
                                       
                                       return(false);
                                    };
                                    
      bool                          Set(int Index,CMorphism &Value)
                                    {
                                       if
                                       (
                                       Index>=0 && Index<Cardinality() && 
                                       Value.domain>=0 && Value.domain<domain.Cardinality() &&
                                       Value.codomain>=0 && Value.codomain<codomain.Cardinality()
                                       )
                                       {
                                          if(!MorphismMatch(Index,Value,morphism,Cardinality()))
                                          {
                                             morphism[Index]=Value;
                                             return(true);
                                          }
                                          
                                       }
                                       
                                       return(false);
                                    };
                                    
      void                          Init(CDomain &Domain,CDomain &Codomain)
                                    {
                                       domain=Domain;
                                       codomain=Codomain;
                                       
                                       init=true;
                                    }
      
      
                                    CHomomorphism(void){ init=false; cardinal=0; };
                                    ~CHomomorphism(void){};
   };
//+------------------------------------------------------------------+
//| ISO-MORPHISM CLASS                                               |
//| For enumerating isomorphisms                                     |
//+------------------------------------------------------------------+
class CIsomorphism
   {
      public:
      
      CMorphism                     morphisms[];
      
                                    CIsomorphism()
                                    {
                                       ArrayFree(morphisms);
                                    };
                                    ~CIsomorphism(){};
   };
//+------------------------------------------------------------------+
//| CATEGORY CLASS                                                   |
//+------------------------------------------------------------------+
class CCategory
   {
      protected:
      
      int                           domains;
      int                           homomorphisms;
      
      CDomain                       domain[];
      CHomomorphism                 homomorphism[];
      
      public:
      
      int                           Domain(){ return(domains); };
      bool                          Domain(int Value){ if(Value>=0 && Value<INT_MAX){ domains=Value; ArrayResize(domain,domains); return(true); } return(false); };
                                    
      int                           Homomorphism(){ return(homomorphisms); };
      bool                          Homomorphism(int Value){ if(Value>=0 && Value<INT_MAX){ homomorphisms=Value; ArrayResize(homomorphism,homomorphisms); return(true); } return(false); };
      
      bool                          Get(int Index,CDomain &D)
                                    {
                                       if(Index>=0 && Index<Domain())
                                       {
                                          D=domain[Index];
                                          
                                          return(true);
                                       }
                                       
                                       return(false);
                                    };
                                    
      bool                          Set(int Index,CDomain &Value)
                                    {
                                       if
                                       (
                                       Index>=0 && Index<Domain()
                                       )
                                       {
                                          domain[Index]=Value;
                                          return(true);
                                       }
                                       
                                       return(false);
                                    };
      
      bool                          Get(int Index,CHomomorphism &H)
                                    {
                                       if(Index>=0 && Index<Homomorphism())
                                       {
                                          H=homomorphism[Index];
                                          
                                          return(true);
                                       }
                                       
                                       return(false);
                                    };
                                    
      bool                          Set(int Index,CHomomorphism &Value)
                                    {
                                       if
                                       (
                                       Index>=0 && Index<Homomorphism()
                                       )
                                       {
                                          homomorphism[Index]=Value;
                                          return(true);
                                       }
                                       
                                       return(false);
                                    };
   };
//+------------------------------------------------------------------+
//| COMMUTATION CLASS                                                |
//+------------------------------------------------------------------+
class CCommutation
   {
      protected:
      
      int                           domains;
      
      CCategory                     category;
      CDomain                       domain[];
      CHomomorphism                 commute;
      
      public:
      
      
   };
//+------------------------------------------------------------------+
//| INPUTS                                                           |
//+------------------------------------------------------------------+
input int __domain_elements=3;
input int __domain_morphisms=5;

//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
   {
      //Declare a sets of natural even & odd numbers
      CDomain _evens,_odds;
      CreateNumberDomain(_evens,__domain_elements,2);
      CreateNumberDomain(_odds,__domain_elements,2,1);
      
      printf(__FUNCSIG__+" evens are... "+PrintDomain(_evens));
      
      printf(__FUNCSIG__+" odds are... "+PrintDomain(_odds));
    
      int _subdomain_count=0; CDomain _subdomains[];
      GetSubdomains(_evens,_subdomain_count,_subdomains);
      //printf(__FUNCSIG__+" evens subs are... "+IntegerToString(_subdomain_count));
      for(int s=0; s<_subdomain_count; s++)
      {
         //printf(" with: "+PrintDomain(_subdomains[s])+", at: "+IntegerToString(s+1));
      }
      
      CDomain _smaller_evens;
      CreateNumberDomain(_smaller_evens,__domain_elements-2,2);
      
      //printf(__FUNCSIG__+" smaller evens are... "+PrintDomain(_smaller_evens));
      
      bool _is_subdomain=IsSubdomain(_evens,_smaller_evens);printf(__FUNCSIG__+" it is: "+string(_is_subdomain)+" that 'smaller-evens' is a subdomain of evens. ");
      _is_subdomain=IsSubdomain(_odds,_smaller_evens);printf(__FUNCSIG__+" it is: "+string(_is_subdomain)+" that 'smaller-evens' is a subdomain of odds. ");
      
      //Declare sets to store time & MA values
      CDomain _time,_close;
      
      MqlRates _rates[];
      
      //Fill domain with 5 most recent bar time & MA values from chart
      if(CopyRates(_Symbol,_Period,0,__domain_morphisms,_rates)>=__domain_morphisms && _time.Cardinality(__domain_morphisms) && _close.Cardinality(__domain_morphisms))
      {
         for(int m=0;m<__domain_morphisms;m++)
         {
            //Create uni row element
            CElement _t,_m;
            if(_t.Cardinality(1) && _m.Cardinality(1))
            {
               datetime _t_value=_rates[m].time;//iTime(_Symbol,_Period,m);
               _t.Set(0,double(int(_t_value)));
               _time.Set(m,_t);
               
               double _m_value=_rates[m].close;//iClose(_Symbol,_Period,5);//,m,MODE_SMA,PRICE_CLOSE);
               _m.Set(0,_m_value);
               _close.Set(m,_m);
            }
         }
      }
      
      //Create homomorphism from time to close
      CHomomorphism _h;_h.Init(_time,_close);
      
      if(_h.init)
      {
         //Create 1-1 morphisms from time to MA
         int _morphisms=0;
         for(int m=0;m<__domain_morphisms;m++)
         {
            if(_h.Cardinality(m,m)){ _morphisms++; }
         }
         
         if(_morphisms>=__domain_morphisms)
         { 
            //printf(__FUNCSIG__+" homomorphism: "+PrintHomomorphism(_h,_Digits,true)); 
         }
      }
      
      //Create homomorphism from time to close
      CHomomorphism _h_image;_h_image.Init(_time,_close);
      
      if(_h_image.init)
      {
         //Create 1-1 morphisms from time to MA
         int _morphisms=0;
         for(int m=0;m<__domain_morphisms;m++)
         {
            int _random_codomain=MathRand()%__domain_morphisms;
            if(_h_image.Cardinality(m,_random_codomain)){ _morphisms++; }
         }
         
         if(_morphisms>=__domain_morphisms)
         { 
            CDomain _image;_image.Cardinality(0);
            Image(_h_image,_image);
            //printf(__FUNCSIG__+" image from homomorphism: "+PrintDomain(_image,_Digits)); 
         }
      }
      
      //
      
      CIsomorphism _i[];
      //is evens isomorphic to odds?
      if(IsIsomorphic(_evens,_odds,_i))
      {
         printf(__FUNCSIG__+" evens is isomorphic to odds by: "+IntegerToString(ArraySize(_i))+" isomorphisms. These are... ");
         for(int s=0; s<ArraySize(_i); s++)
         {
            string _print="";
            for(int ss=0; ss<ArraySize(_i[s].morphisms); ss++)
            {
               _print+=PrintMorphism(_i[s].morphisms[ss],_evens,_odds);
            }
            
            printf(_print+" at: "+IntegerToString(s));
         }
      }
      
      CDomain _ma;
      double _ma_buffer[];
      int _ma_handle=iMA(_Symbol,_Period,__domain_morphisms,0,MODE_SMA,PRICE_CLOSE);
      if(CopyBuffer(_ma_handle,0,0,__domain_morphisms,_ma_buffer)>=__domain_morphisms && _ma.Cardinality(__domain_morphisms))
      {
         for(int e=0;e<__domain_morphisms;e++)
         {
            //Create uni row element
            CElement _e;
            if(_e.Cardinality(1))
            {
               double _ma_value=_ma_buffer[e];
               _e.Set(0,_ma_value);
               _ma.Set(e,_e);
            }
         } 
         
         //Create homomorphism from time to close
         CHomomorphism _ma_image;_ma_image.Init(_ma,_time);
         
         if(_ma_image.init)
         {
            //Create 1-1 morphisms from time to MA
            int _morphisms=0;
            for(int m=0;m<__domain_morphisms;m++)
            {
               int _random_codomain=MathRand()%__domain_morphisms;
               if(_ma_image.Cardinality(m,_random_codomain)){ _morphisms++; }
            }
            printf(__FUNCSIG__+"  ma init morphs: "+IntegerToString(_morphisms));
            
            if(_morphisms>=__domain_morphisms)
            { 
               CDomain _image;_image.Cardinality(0);
               Image(_ma_image,_image);
               printf(__FUNCSIG__+" image from ma homomorphism: "+PrintDomain(_image,0,true)); 
            }
         }
      }
   }
//+------------------------------------------------------------------+
//| Create Simple Domain(Set) of natural numbers                     |
//+------------------------------------------------------------------+
void CreateNumberDomain(CDomain &S,int SetSize,int RemainderType,int Remainder=0)
   {
      //Set its cardinal to input value
      S.Cardinality(SetSize);
      
      //Assign numbers to the domain in ascending order
      int _assigned=0,_number=0;
      while(_assigned<SetSize)
      {
         if(_number%RemainderType==Remainder)
         {
            CElement _e;
            _e.Cardinality(1);
            if(_e.Set(0,_number) && S.Set(_assigned,_e,true))//check to ensure element is assigned
            {
               _assigned++;
               
               if(_assigned>=SetSize)
               {
                  break;
               }
            }
         }
         _number++;
      }
   }
//+------------------------------------------------------------------+
//| Print Element function                                            |
//+------------------------------------------------------------------+
string PrintElement(CElement &O,int Precision=1,bool IsTime=false)
   {
      string _element="(";
      //
      for(int r=0; r<O.Cardinality(); r++)
      {
         if(!IsTime)
         {
            _element+=DoubleToString(O.Get(r),Precision);
         }
         else if(IsTime)
         {
            _element+=TimeToString(datetime(int(O.Get(r))));
         }
         
         if(r<O.Cardinality()-1){ _element+=","; }
      }
      //
      return(_element+")");
   }
//+------------------------------------------------------------------+
//| Print Set function                                               |
//+------------------------------------------------------------------+
string PrintDomain(CDomain &S,int Precision=1,bool IsTime=false)
   {
      string _set="{";
      //
      CElement _e;
      for(int o=0; o<S.Cardinality(); o++)
      {
         S.Get(o,_e);
         _set+=PrintElement(_e,Precision,IsTime);if(o<S.Cardinality()-1){ _set+=","; }
      }
      //
      return(_set+"}\n");
   }
//+------------------------------------------------------------------+
//| Print Morphism function                                          |
//+------------------------------------------------------------------+
string PrintMorphism(CMorphism &M, CDomain &Domain,CDomain &Codomain,int Precision=1,bool DomainIsTime=false,bool CodomainIsTime=false)
   {
      string _morphism="";
      //
      CElement _d,_c;
      if(Domain.Get(M.domain,_d) && Codomain.Get(M.codomain,_c))
      {
         _morphism=PrintElement(_d,Precision,DomainIsTime);
         _morphism+="|----->";
         _morphism+=PrintElement(_c,Precision,CodomainIsTime);
         _morphism+="\n";
      }
      //
      return(_morphism);
   }
//+------------------------------------------------------------------+
//| Print Homomorphism function                                      |
//+------------------------------------------------------------------+
string PrintHomomorphism(CHomomorphism &H,int Precision=1,bool DomainIsTime=false,bool CodomainIsTime=false)
   {
      string _homomorphism="\n\n"+PrintDomain(H.domain,Precision,DomainIsTime);
      //
      _homomorphism+="|\n";
      
      CMorphism _m;
      for(int m=0;m<H.Cardinality();m++)
      {
         if(H.Get(m,_m))
         {
            _homomorphism+=(PrintMorphism(_m,H.domain,H.codomain,Precision,DomainIsTime,CodomainIsTime));
         }
      }
      //
      _homomorphism+="|\n";
      
      _homomorphism+=PrintDomain(H.codomain,Precision,CodomainIsTime);
      //
      return(_homomorphism);
   }
//+------------------------------------------------------------------+
//| Get Subdomains function                                             |
//+------------------------------------------------------------------+
void GetSubdomains(CDomain &S,int &SubdomainCount,CDomain &Subdomains[])
   {
      for(int o=1; o<=S.Cardinality(); o++)
      {
         GetSubdomains(S,0,0,o,SubdomainCount,Subdomains);
      }
   }
//+------------------------------------------------------------------+
//| Get Subdomains function                                             |
//+------------------------------------------------------------------+
void GetSubdomains(CDomain &S,int Start,int Index,int Number,int &SubdomainCount,CDomain &Subdomains[])
   {
      // Get domain cardinal
      int _n = S.Cardinality();
      
      int _i,_j;
      
      if(Index-Start+1==Number)
      {
         if(Number==1)
         {
            CElement _e;
            
            if(_e.Cardinality(0))
            {
               for(_i=0; _i<_n; _i++)
               {
                  SubdomainCount++;
                  ArrayResize(Subdomains,SubdomainCount);
                  
                  if(ArraySize(Subdomains)==SubdomainCount && Subdomains[SubdomainCount-1].Cardinality(0) && S.Get(_i,_e) && Subdomains[SubdomainCount-1].Cardinality(1))
                  {
                     Subdomains[SubdomainCount-1].Set(0,_e,true);
                  }
               }
            }
         }
         else
         {
            for(_j=Index; _j<_n; _j++)
            {
               SubdomainCount++;
               ArrayResize(Subdomains,SubdomainCount);
               
               CElement _e;
               
               if(_e.Cardinality(0) && ArraySize(Subdomains)==SubdomainCount && Subdomains[SubdomainCount-1].Cardinality(0))
               {
                  for(_i=Start; _i<Index; _i++)
                  {
                     if(S.Get(_i,_e) && Subdomains[SubdomainCount-1].Cardinality(Subdomains[SubdomainCount-1].Cardinality()+1))
                     {
                        Subdomains[SubdomainCount-1].Set(Subdomains[SubdomainCount-1].Cardinality()-1,_e,true);
                     }
                  }
                  
                  if(S.Get(_j,_e) && Subdomains[SubdomainCount-1].Cardinality(Subdomains[SubdomainCount-1].Cardinality()+1))
                  {
                     Subdomains[SubdomainCount-1].Set(Subdomains[SubdomainCount-1].Cardinality()-1,_e,true);
                  }
               }
            }
            
            if(Start!=_n-Number)
            {
               GetSubdomains(S,Start+1,Start+1,Number,SubdomainCount,Subdomains);
            }
         }
      }
      else
      {
         GetSubdomains(S,Start,Index+1,Number,SubdomainCount,Subdomains);
      }
   }
//+------------------------------------------------------------------+
//| Element Match function                                            |
//+------------------------------------------------------------------+
bool ElementMatch(CElement &A,CElement &B)
   {
      if(A.Cardinality()!=B.Cardinality())
      {
         return(false);
      }
      
      bool _matched=true;
      
      for(int r=0; r<A.Cardinality(); r++)
      {
         if(A.Get(r)!=B.Get(r))
         {
            _matched=false;
            break;
         }
      }
      
      return(_matched);
   }
//+------------------------------------------------------------------+
//| Set Match function                                               |
//+------------------------------------------------------------------+
bool DomainMatch(CDomain &A,CDomain &B)
   {
      if(A.Cardinality()!=B.Cardinality())
      {
         return(false);
      }
      
      bool _matched=true;
      
      for(int o=0; o<A.Cardinality(); o++)
      {
         CElement _a,_b;
         
         if(A.Get(o,_a) && B.Get(o,_b) && !ElementMatch(_a,_b))
         {
            _matched=false; break;
         }
      }
      
      return(_matched);
   }
//+------------------------------------------------------------------+
//| Is Subdomain function                                               |
//+------------------------------------------------------------------+
bool IsSubdomain(CDomain &Domain,CDomain &SubDomain)
   {
      bool _is_subdomain=false;
      
      int _subdomain_count=0; CDomain _subdomains[];
      GetSubdomains(Domain,_subdomain_count,_subdomains);
      
      for(int c=0;c<_subdomain_count;c++)
      {
         if(DomainMatch(SubDomain,_subdomains[c]))
         {
            _is_subdomain=true;
            break;
         }
      }
      
      return(_is_subdomain);
   }
//+------------------------------------------------------------------+
//| Morphism Match function                                          |
//+------------------------------------------------------------------+
bool MorphismMatch(int Index,CMorphism &Value,CMorphism &Morphism[],int Cardinality)
   {
      if(Cardinality<0)
      {
         return(true);
      }
      
      bool _matched=true;
      
      for(int m=0; m<Cardinality; m++)
      {
         if(m==Index){ continue; }
         
         if(Morphism[m].domain==Value.domain)
         {
            _matched=false; break;
         }
      }
      
      return(_matched);
   }
//+------------------------------------------------------------------+
//| Image function                                                   |
//+------------------------------------------------------------------+
void Image(CHomomorphism &H,CDomain &Output)
   {
      for(int m=0;m<H.Cardinality();m++)
      {
         CElement _e;
         CMorphism _m;
         if(H.Get(m,_m) && H.codomain.Get(_m.codomain,_e))
         {
            bool _matched=false;
            for(int o=0;o<Output.Cardinality();o++)
            {
               CElement _ee;
               if(Output.Get(o,_ee) && ElementMatch(_e,_ee))
               {
                  _matched=true; break;
               }
            }
            
            if(!_matched)
            {
               Output.Cardinality(Output.Cardinality()+1);
               Output.Set(Output.Cardinality()-1,_e,true);
            }
         }
      }
   }
//+------------------------------------------------------------------+
//| Factorial function                                               |
//+------------------------------------------------------------------+
uint MathFactorial(uint N)
   {
      uint _factorial=N,_n=N-1;
      
      while(_n>0)
      {
         _factorial*=_n;
         
         _n--;
         
         if(_n<=0){ break; }
      }
      
      return(_factorial);
   }
//+------------------------------------------------------------------+
//| Get Isomorphisms function                                        |
//+------------------------------------------------------------------+
bool IsIsomorphic(CDomain &A,CDomain &B,CIsomorphism &Output[])
   {
      if(A.Cardinality()!=B.Cardinality())
      {
         return(false);
      }
      
      uint _cardinality=MathFactorial(A.Cardinality());
      
      ArrayResize(Output,_cardinality);
      
      for(uint s=0;s<_cardinality;s++)
      {
         ArrayResize(Output[s].morphisms,A.Cardinality());
      }
      
      int _index=0;
      CDomain _output[];ArrayResize(_output,_cardinality);
      GetIsomorphisms(B, 0, A.Cardinality()-1, A.Cardinality(), _index, _output);
      
      for(uint s=0;s<_cardinality;s++)
      {
         for(int ss=0;ss<A.Cardinality();ss++)
         {
            CElement _eo;
            if(_output[s].Get(ss,_eo))
            {
               for(int sss=0;sss<B.Cardinality();sss++)
               {
                  CElement _eb;
                  if(B.Get(sss,_eb))
                  {
                     if(ElementMatch(_eo,_eb))
                     {
                        Output[s].morphisms[ss].codomain=sss;
                        //break;
                     }
                  }
               }
            }
            
            Output[s].morphisms[ss].domain=ss;
         }
      }
      
      return(true);
   }
//+------------------------------------------------------------------+
//| Get Isomorphisms function                                        |
//+------------------------------------------------------------------+
void GetIsomorphisms(CDomain &B,int Start,int Stop,int Cardinality,int &Index,CDomain &Output[]) 
   {
      if(Start == Stop) 
      {
         Output[Index].Cardinality(0);
         Output[Index].Cardinality(B.Cardinality());
         // the permutation
         for (int i = 0; i <= Stop; i++) 
         {
            CElement _e;
            if(B.Get(i,_e))
            {
               Output[Index].Set(i,_e,true);
            }
         }
         
         Index++;
      } 
      else 
      {
         CElement _e;
         if(B.Get(Start,_e))
         {
            for (int i = Start; i <= Stop; i++) 
            {
               // swap the elements at indices start and i
               CElement _ei;
               if(B.Get(i,_ei))
               {
                  CElement _temp = _e;
                  if(B.Set(Start,_ei))
                  {
                     if(B.Set(i,_e))
                     {
                        // recursively permute the remaining elements
                        GetIsomorphisms(B, Start + 1, Stop, Cardinality, Index, Output);
                     
                        // swap the elements back to their original positions
                        if(B.Set(Start,_e))
                        {
                           B.Set(i,_ei);
                        }
                     }
                  }
               }
            }
         }
      }
   }
//+------------------------------------------------------------------+
